#!/usr/local/bin/python

# Copyright (c) 2007 The Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions, and the following disclaimer.
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions, and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#   3. Redistributions must acknowledge that this software was
#      originally developed by the UCSF Computer Graphics Laboratory
#      under support by the NIH National Center for Research Resources,
#      grant P41-RR01081.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from xmllib import XMLParser
from PIL import Image, ImageDraw, ImageFont

def geneImageMap(xmlFile,topoSnp):
	topo = GeneTopo(xmlFile)
	return topo.imageMap(topoSnp)

class GeneTopo:
	def __init__(self, filename):
		self.name = filename
		self.locations = {}
		f = open(filename)
		data = f.read()
		f.close()
		TopoParser().topoParseString(self, data)
		self.symbolMap = {
			"0":		self._symbolNone,
			"none":		self._symbolNone,
			"1":		self._symbolCircle,
			"circle":	self._symbolCircle,
			"2":		self._symbolSquare,
			"square":	self._symbolSquare,
			"3":		self._symbolHexagon,
			"hexagon":	self._symbolHexagon,
			"4":		self._symbolUparrow,
			"uparrow":	self._symbolUparrow,
			"5":		self._symbolDownarrow,
			"downarrow":	self._symbolDownarrow,
		}
		self.aaFont = ImageFont.load_path("PIL/fonts/clR4x6.pil")
		self.labelFont = ImageFont.load_path("PIL/fonts/clB8x16.pil")
		self.size = self._getSize()
		self.scale = self._getScale(self.size)
		self._setTop()

	def _getSize(self):
		return max(self.aaFont.getsize("M")) + 2

	def _getScale(self, size):
		return self.size / 2.0 / self.radius

	def _setTop(self):
		if self.labelTop:
			self.topSize = self.labelFont.getsize(self.labelTop)
			self.topHt = self.topSize[1] + 4	# 2 pixel border
		else:
			self.topSize = (0, 0)
			self.topHt = 0

	def getImageMap(self,topoSnp):

		if self.labelBottom:
			bottomSize = self.labelFont.getsize(self.labelBottom)
			self.bottomHt = bottomSize[1] + 4
		else:
			bottomSize = (0, 0)
			self.bottomHt = 0

		bboxWidth = self.xMax - self.xMin + 2*self.radius
		bboxHeight = self.yMax - self.yMin
		width = bboxWidth * self.scale
		height = bboxHeight * self.scale + self.topHt + self.bottomHt

		img = Image.new("RGB", (width, height), 0xFFFFFF)
		draw = ImageDraw.Draw(img)

		# Render top and bottom labels
		if self.labelTop:
			draw.text(((width - self.topSize[0]) / 2, 2),
					self.labelTop,
					font=self.labelFont, fill=0)
		if self.labelBottom:
			draw.text(((width - bottomSize[0]) / 2,
					height - self.bottomHt + 2),
					self.labelBottom,
					font=self.labelFont, fill=0)

		# Render membrane borders
		mLeft, mTop = self.mapXY(self.xMembMin, self.yMembTop)
		mRight, mBottom = self.mapXY(self.xMembMax, self.yMembBottom)
		draw.line([mLeft, mTop, mRight, mTop], fill=0x0)
		draw.line([mLeft, mBottom, mRight, mBottom], fill=0x0)

		# Render each amino acid
		for loc in self.locations.values():
			x, y = self.mapXY(loc.x, loc.y + self.radius)
			renderer = self.symbolMap.get(loc.symbol,
							self._symbolCircle)
			renderer(draw, x, y, self.size,
					loc.background, loc.border)
			w, h = self.aaFont.getsize(loc.type)
			x = x + (self.size - w) / 2 + 1
			y = y + (self.size - h)
			draw.text((x, y), loc.type, font=self.aaFont,
					fill=loc.foreground)


	        # Zero bridges
        	#
        	try:
                	zbList = topo.zerobridge
        	except AttributeError:
                	pass
        	else:
                	for zb in zbList:
				longs = len(zbList)
				#print "there are %d zero bridges" %longs
				j = 1
				while j <= longs:
					zbpts = zbList[j-1]
					x1,y1 = zbpts[0]	
					x2,y2 = zbpts[1]	
					x3,y3 = zbpts[2]	
					x4,y4 = zbpts[3]	
					x1pt,y1pt = self.mapXY(x1,y1)
					x2pt,y2pt = self.mapXY(x2,y2)
					x3pt,y3pt = self.mapXY(x3,y3)
					x4pt,y4pt = self.mapXY(x4,y4)
					draw.line([x1pt,y1pt,x2pt,y2pt], fill=0x0)
					draw.line([x2pt,y2pt,x3pt,y3pt], fill=0x0)
					draw.line([x3pt,y3pt,x4pt,y4pt], fill=0x0)
					j = j + 1

		return img

	def mapXY(self, x, y):
		nx = int((x - self.xMin) * self.scale)
		ny = int((self.yMax - y) * self.scale) + self.topHt
		return nx, ny

	def _symbolNone(self, draw, x, y, size, bg, border):
		pass

	def _symbolCircle(self, draw, x, y, size, bg, border):
		bbox = (x, y, x + size, y + size)
		draw.ellipse(bbox, fill=bg, outline=border)

	def _symbolSquare(self, draw, x, y, size, bg, border):
		bbox = (x, y, x + size, y + size)
		draw.rectangle(bbox, fill=bg, outline=border)

	def _symbolHexagon(self, draw, x, y, size, bg, border):
		draw.polygon(hexagon(x, y, size, 0), fill=bg, outline=border)

	def _symbolUparrow(self, draw, x, y, size, bg, border):
		draw.polygon(uparrow(x, y, size, 0), fill=bg, outline=border)

	def _symbolDownarrow(self, draw, x, y, size, bg, border):
		draw.polygon(downarrow(x, y, size, 0), fill=bg, outline=border)


def diamond(x, y, size, hlWidth):
	# small diamonds are hard to see, so give them a boost
	#x -= 1; y -= 1; size += 2
	radius = size / 2.0
	cx = x + radius
	cy = y + radius
	radius = radius + hlWidth
	dx = radius
	dy = radius * 0.50
	return ((cx, cy + radius), (cx + dx, cy),
		(cx, cy - radius), (cx - dx, cy))

def hexagon(x, y, size, hlWidth):
	radius = size / 2.0
	cx = x + radius
	cy = y + radius
	radius = radius + hlWidth
	dx = radius
	dy = radius / 2
	return ((cx, cy + radius), (cx + dx, cy + dy), (cx + dx, cy - dy),
		(cx, cy - radius), (cx - dx, cy - dy), (cx - dx, cy + dy))

def uparrow(x, y, size, hlWidth):
	radius = size / 2.0
	cx = x + radius
	cy = y + radius
	radius = radius + hlWidth
	dx = radius
	dy = radius * 0.50
	return ((cx, cy + radius), (cx + dx, cy + dy),
		(cx, cy - radius), (cx - dx, cy + dy))

def downarrow(x, y, size, hlWidth):
	radius = size / 2.0
	cx = x + radius
	cy = y + radius
	radius = radius + hlWidth
	dx = radius
	dy = radius * 0.50
	return ((cx, cy + radius), (cx + dx, cy - dy),
		(cx, cy - radius), (cx - dx, cy - dy))

class Location:
	pass

class TopoParser(XMLParser):

	def __init__(self, *args, **kw):
		apply(XMLParser.__init__, (self,) + args, kw)
		self.attributes = {
			"TOPO":	{
				"name":		None,
			},
			"bbox":	{
				"minx":		None,
				"maxx":		None,
				"miny":		None,
				"maxy":		None,
			},
			"layout": {
				"radius":	"5.0",
				"label_top":	"",
				"label_bot":	"",
			},
			"membrane":	{
				"m_minx":	None,
				"m_maxx":	None,
				"m_topy":	None,
				"m_boty":	None,
			},
                        "zerobridge":   {
                                "b_start_x":    None,
                                "b_start_y":    None,
                                "b_ltop_x":     None,
                                "b_ltop_y":     None,
                                "b_rtop_x":     None,
                                "b_rtop_y":     None,
                                "b_end_x":      None,
                                "b_end_y":      None,
                        },
			"entry":	{
				"position":	None,
				"residue":	None,
				"x":		None,
				"y":		None,
				"symbol":	"1",
				"foreground":	"0 0 0",
				"background":	"1 1 1",
				"border":	"0 0 0",
			},
		}
		self.elements = {
			"TOPO": (self.topo_start, self.topo_end),
			"bbox": (self.bbox_start, self.bbox_end),
			"layout": (self.layout_start, self.layout_end),
			"membrane": (self.membrane_start, self.membrane_end),
			"zerobridge": (self.zerobridge_start, self.zerobridge_end),
			"entry": (self.entry_start, self.entry_end),
		}

	def topoParseString(self, topo, data):
		self.topo = topo
		self.reset()
		self.feed(data)
		self.close()
		del self.topo

	def topo_start(self, attr):
		self.topo.gene = attr["name"]
		
	def topo_end(self):
		pass

	def bbox_start(self, attr):
		self.topo.xMin = float(attr["minx"])
		self.topo.xMax = float(attr["maxx"])
		self.topo.yMin = float(attr["miny"])
		self.topo.yMax = float(attr["maxy"])
	def bbox_end(self):
		pass

	def layout_start(self, attr):
		self.topo.radius = float(attr["radius"])
		self.topo.labelTop = attr["label_top"]
		self.topo.labelBottom = attr["label_bot"]
	def layout_end(self):
		pass

	def membrane_start(self, attr):
		self.topo.xMembMin = float(attr["m_minx"])
		self.topo.xMembMax = float(attr["m_maxx"])
		self.topo.yMembTop = float(attr["m_topy"])
		self.topo.yMembBottom = float(attr["m_boty"])
	def membrane_end(self):
		pass

        def zerobridge_start(self, attr):
                points = [
                        ( float(attr["b_start_x"]), float(attr["b_start_y"]) ),
                        ( float(attr["b_ltop_x"]), float(attr["b_ltop_y"]) ),
                        ( float(attr["b_rtop_x"]), float(attr["b_rtop_y"]) ),
                        ( float(attr["b_end_x"]), float(attr["b_end_y"]) ),
                ]
                try:
                        self.topo.zerobridge.append(points)
                except AttributeError:
                        self.topo.zerobridge = [ points ]
        def zerobridge_end(self):
                pass

 	def entry_start(self, attr):
		loc = Location()
		loc.sequence = int(attr["position"])
		loc.type = attr["residue"]
		loc.x = float(attr["x"])
		loc.y = float(attr["y"])
		loc.symbol = attr["symbol"]
		loc.foreground = rgb2pil(str2rgb(attr["foreground"]))
		loc.background = rgb2pil(str2rgb(attr["background"]))
		loc.border = rgb2pil(str2rgb(attr["border"]))
		self.topo.locations[loc.sequence] = loc
	def entry_end(self):
		pass

def str2rgb(s):
	return map(float, s.split())

def rgb2str(rgb):
	return "%g %g %g" % tuple(rgb)

def rgb2pil(rgb):
	return ((int(rgb[2] * 255) << 16)
		| (int(rgb[1] * 255) << 8) | int(rgb[0] * 255))

if __name__ == "__main__":
	import sys

	if len(sys.argv) > 1:
		filename = sys.argv[1]
	else:
		filename = "topo.xml"

	topoSnp = {}
	topo = GeneTopo(filename)
	img = topo.getImageMap(topoSnp)

	try:
		img.save("test.png", "PNG", optimize=1)
	except IOError:
		img.save("test.gif", "GIF")

